﻿using System;
using System.Data.SQLite;
using System.Text;
using System.Collections.Generic;
using PhoneBook.Extensions;

namespace PhoneBook
{
    namespace Model
    {
        public class Model : IDisposable
        {
            public event EventHandler<ContactAddedEventArg> ContactAddedEvent;
            public event EventHandler<ContactRemovedEventArg> ContactRemovedEvent;

            public Model()
            {
                if (!System.IO.File.Exists(dbName_))
                {
                    throw new ModelException("Database file was not found.");
                }

                connection_ = new SQLiteConnection("Data Source = " + dbName_);
                connection_.Open();
            }

            public void AddContact(Contact contact)
            {
                ThrowExceptionIfDisposed();

                try
                {
                    var query = new StringBuilder();

                    query.AppendFormat(
                        "INSERT INTO Accounts(first_name, last_name, phone, image) VALUES('{0}', '{1}', '{2}', @image);",
                        contact.FirstName,
                        contact.LastName,
                        contact.Phone
                        );

                    using (SQLiteCommand command = new SQLiteCommand(connection_))
                    {
                        command.CommandText = query.ToString();
                        command.CommandType = System.Data.CommandType.Text;

                        byte[] img = contact.Image == null ? null : contact.Image.ToByteArray();
                        command.Parameters.Add("@image", System.Data.DbType.Binary, img == null ? 0 : img.Length).Value = img;

                        using (SQLiteTransaction transaction = connection_.BeginTransaction())
                        {
                            command.ExecuteScalar();
                            var rowID = connection_.LastInsertRowId;
                            contact.Id = rowID;
                            transaction.Commit();
                        }
                    }

                    ContactAddedEvent?.Invoke(this, new ContactAddedEventArg(contact));
                }
                catch (Exception e)
                {
                    throw new ModelException(e.Message, e);
                }
            }

            public void RemoveContact(long id)
            {
                ThrowExceptionIfDisposed();

                try
                {
                    var query = new StringBuilder();
                    query.AppendFormat("DELETE FROM Accounts WHERE id = '{0}'", id);

                    using (SQLiteCommand command = new SQLiteCommand(connection_))
                    {
                        command.CommandText = query.ToString();
                        command.CommandType = System.Data.CommandType.Text;
                        var execResult = command.ExecuteNonQuery();

                        if (execResult == -1)
                        {
                            throw new InvalidOperationException("Contact was not added!");
                        }
                    }

                    ContactRemovedEvent?.Invoke(this, new ContactRemovedEventArg(id));
                }
                catch (Exception e)
                {
                    throw new ModelException(e.Message, e);
                }
            }

            public long ContactCount
            {
                get
                {
                    ThrowExceptionIfDisposed();

                    try
                    {
                        string query = "SELECT count(id) FROM ACCOUNTS";
                        var reader = ExecuteReader(query, System.Data.CommandBehavior.SingleRow);

                        if (reader.Read())
                        {
                            return reader.GetInt64(0);
                        }

                        return 0;
                    }
                    catch (Exception e)
                    {
                        throw new ModelException(e.Message, e);
                    }
                }
            }

            public Contact SelectContact(long id)
            {
                ThrowExceptionIfDisposed();

                try
                {
                    var query = new StringBuilder();
                    query.AppendFormat("SELECT * FROM ACCOUNTS WHERE id = {0}", id);

                    using (var reader = ExecuteReader(query.ToString(), System.Data.CommandBehavior.Default | System.Data.CommandBehavior.KeyInfo))
                    {
                        if (reader.Read())
                        {
                            return ReadContact(reader);
                        }
                    }

                    return null;
                }
                catch (Exception e)
                {
                    throw new ModelException(e.Message, e);
                }
            }

            public IEnumerable<Contact> SelectContacts(long from, long to)
            {
                ThrowExceptionIfDisposed();

                try
                {
                    if (to < from)
                    {
                        throw new ArgumentException("from must be leq than to");
                    }

                    var query = new StringBuilder();

                    var count = to - from + 1;
                    query.AppendFormat("SELECT * FROM ACCOUNTS LIMIT {0} OFFSET {1}", count, from);

                    var contacts = new LinkedList<Contact>();

                    using (var reader = ExecuteReader(query.ToString(), System.Data.CommandBehavior.Default | System.Data.CommandBehavior.KeyInfo))
                    {
                        while (reader.Read())
                        {
                            contacts.AddLast(ReadContact(reader));
                        }
                    }

                    return contacts;
                }
                catch (Exception e)
                {
                    throw new ModelException(e.Message, e);
                }
            }

            public void Dispose()
            {
                Dispose(true);
                GC.SuppressFinalize(this);
            }

            private void Dispose(bool disposing)
            {
                if (!disposed_)
                {
                    if (disposing)
                    {
                        connection_.Dispose();
                    }

                    disposed_ = true;
                }
            }

            private void ThrowExceptionIfDisposed()
            {
                if (disposed_)
                {
                    throw new ObjectDisposedException("The model is disposed");
                }
            }

            private static Contact ReadContact(SQLiteDataReader reader)
            {
                System.Drawing.Image img = null;

                if (!reader.IsDBNull(imageCol))
                {
                    using (var blob = reader.GetBlob(imageCol, true))
                    {
                        img = blob.ToImage();
                    }
                }

                return new Contact
                {
                    Id = reader.GetInt64(contactId),
                    FirstName = reader.GetString(firstNameCol),
                    LastName = reader.GetString(lastNameCol),
                    Phone = reader.GetString(phoneCol),
                    Image = img
                };
            }

            private SQLiteDataReader ExecuteReader(string query, System.Data.CommandBehavior commandBehavior)
            {
                using (SQLiteCommand command = new SQLiteCommand(connection_))
                {
                    command.CommandText = query;
                    command.CommandType = System.Data.CommandType.Text;

                    return command.ExecuteReader(commandBehavior);
                }
            }

            private const string dbName_ = "Contacts.db";

            private const int contactId = 0;
            private const int firstNameCol = 1;
            private const int lastNameCol = 2;
            private const int phoneCol = 3;
            private const int imageCol = 4;

            private bool disposed_;
            private SQLiteConnection connection_;
        }
    }
}
